@@ -12,14 +12,14 @@
 
 use crate::factor::{pollard_rho, trial_division};
 use crate::nt_funcs::{
-    factors64, is_prime64, next_prime, nth_prime_bounds, nth_prime_est, prev_prime,
+    factorize128, factors, is_prime64, next_prime, nth_prime_bounds, nth_prime_est, prev_prime,
 };
 use crate::primality::{PrimalityBase, PrimalityRefBase};
 use crate::tables::{SMALL_PRIMES, SMALL_PRIMES_NEXT};
 use crate::traits::{
     FactorizationConfig, Primality, PrimalityTestConfig, PrimalityUtils, PrimeBuffer,
 };
-use bitvec::bitvec;
+use bitvec::{bitvec, prelude::Msb0};
 use lru::LruCache;
 use num_integer::Roots;
 use rand::random;
@@ -27,8 +27,14 @@ use std::collections::BTreeMap;
 
 /// Extension functions that can utilize pre-generated primes
 pub trait PrimeBufferExt: for<'a> PrimeBuffer<'a> {
-    /// Test if an integer is a prime. The `config` will take effect only if the target is larger
-    /// than 2^64, otherwise [is_prime64] will be used.
+    /// Test if an integer is a prime.
+    ///
+    /// For targets smaller than 2^64, the deterministic [is_prime64] will be used, otherwise
+    /// the primality test algorithms can be specified by the `config` argument.
+    ///
+    /// The primality test can be either deterministic or probabilistic for large integers depending on the `config`.
+    /// The return value is represented by the enum [Primality], which tells whether the primality test is deterministic
+    /// or probabilistic.
     fn is_prime<T: PrimalityBase>(
         &self,
         target: &T,
@@ -75,8 +81,8 @@ pub trait PrimeBufferExt: for<'a> PrimeBuffer<'a> {
             probability *= 1. - 0.25f32.powi(config.sprp_random_trials as i32);
         }
         if !witness_list
-            .iter()
-            .all(|&x| target.is_sprp(T::from_u64(x).unwrap()))
+            .into_iter()
+            .all(|x| target.is_sprp(T::from_u64(x).unwrap()))
         {
             return Primality::No;
         }
@@ -98,11 +104,17 @@ pub trait PrimeBufferExt: for<'a> PrimeBuffer<'a> {
         Primality::Probable(probability)
     }
 
-    /// Factorize an integer. The `config` will take effect only if the target is larger
-    /// than 2^64, otherwise [factors64] will be used.
+    /// Factorize an integer.
     ///
-    /// The factorization result will be returned as a map from primes to exponents. If the
-    /// factorization failed, then a list of found factors will be returned.
+    /// For targets smaller than 2^64, the efficient [factorize64] will be used, otherwise
+    /// the primality test and factorization algorithms can be specified by the `config` argument.
+    ///
+    /// The factorization result will be returned as a map from prime factors to their exponents. If the
+    /// factorization failed, then a list of found factors (not necessarily primes) will be returned. A prime
+    /// factor will repeat if its exponent is larget than one, and it's ensured that the product of the list of
+    /// factors is equal to the original target.
+    /// 
+    /// TODO(v0.next): Return two lists when failed, one for prime factors, another one for remaining cofactors
     fn factors<T: PrimalityBase>(
         &self,
         target: T,
@@ -111,11 +123,11 @@ pub trait PrimeBufferExt: for<'a> PrimeBuffer<'a> {
     where
         for<'r> &'r T: PrimalityRefBase<T>,
     {
-        // shortcut if the target is in u64 range
-        if let Some(x) = target.to_u64() {
-            return Ok(factors64(x)
-                .iter()
-                .map(|(&k, &v)| (T::from_u64(k).unwrap(), v))
+        // shortcut if the target is in u128 range
+        if let Some(x) = target.to_u128() {
+            return Ok(factorize128(x)
+                .into_iter()
+                .map(|(k, v)| (T::from_u128(k).unwrap(), v))
                 .collect());
         }
         let config = config.unwrap_or(FactorizationConfig::default());
@@ -127,6 +139,8 @@ pub trait PrimeBufferExt: for<'a> PrimeBuffer<'a> {
             .map(|(k, v)| (T::from_u64(k).unwrap(), v))
             .collect();
 
+        // TODO: check is_perfect_power before other methods
+
         // find factors by dividing
         let mut successful = true;
         let mut config = config;
@@ -142,7 +156,7 @@ pub trait PrimeBufferExt: for<'a> PrimeBuffer<'a> {
                 let mut todo = vec![res];
                 while let Some(target) = todo.pop() {
                     if self
-                        .is_prime(&target, Some(config.prime_test_config))
+                        .is_prime(&target, Some(config.primality_config))
                         .probably()
                     {
                         *result.entry(target).or_insert(0) += 1;
@@ -169,6 +183,22 @@ pub trait PrimeBufferExt: for<'a> PrimeBuffer<'a> {
         }
     }
 
+    /// Factorize an integer until all prime factors are found.
+    ///
+    /// This function will try to call [factors] function repeatedly until the target
+    /// is fully factorized.
+    fn factorize<T: PrimalityBase>(&self, target: T) -> BTreeMap<T, usize>
+    where
+        for<'r> &'r T: PrimalityRefBase<T>,
+    {
+        // TODO: change to a reasonable strategy after default factorization failed
+        loop {
+            if let Ok(result) = factors(target.clone(), None) {
+                break result;
+            }
+        }
+    }
+
     /// Return a proper divisor of target (randomly), even works for very large numbers.
     /// Return `None` if no factor is found.
     ///
@@ -272,7 +302,7 @@ impl<'a> PrimeBuffer<'a> for NaiveBuffer {
         }
 
         // create sieve and filter with existing primes
-        let mut sieve = bitvec![0; ((sieve_limit - current) / 2) as usize];
+        let mut sieve = bitvec![usize, Msb0; 0; ((sieve_limit - current) / 2) as usize];
         for p in self.list.iter().skip(1) {
             // skip pre-filtered 2
             let start = if p * p < current {
@@ -313,6 +343,10 @@ impl NaiveBuffer {
     }
 
     /// Returns all primes ≤ `limit`. The primes are sorted.
+    // TODO: let primes return &[u64] instead of iterator, and create a separate iterator
+    //       for endless prime iter. This can be a method in this trait, or standalone function,
+    //       or implement as IntoIter. We can try to implement PrimeBuffer on primal first and see
+    //       if it's reasonable to unifiy
     pub fn primes(&mut self, limit: u64) -> std::iter::Take<<Self as PrimeBuffer>::PrimeIter> {
         self.reserve(limit);
         let position = match self.list.binary_search(&limit) {
@@ -445,6 +479,7 @@ impl NaiveBuffer {
 #[cfg(test)]
 mod tests {
     use super::*;
+    use crate::detail::Mint;
     #[cfg(feature = "num-bigint")]
     use core::str::FromStr;
     #[cfg(feature = "num-bigint")]
@@ -506,23 +541,27 @@ mod tests {
     }
 
     #[test]
-    fn pb_is_prime_test() {
+    fn is_prime_test() {
         // test for is_prime
         let pb = NaiveBuffer::new();
 
         // some mersenne numbers
-        assert!(matches!(
+        assert_eq!(
             pb.is_prime(&(2u32.pow(19) - 1), None),
             Primality::Yes
-        ));
-        assert!(matches!(
+        );
+        assert_eq!(
             pb.is_prime(&(2u32.pow(23) - 1), None),
             Primality::No
-        ));
+        );
         assert!(matches!(
             pb.is_prime(&(2u128.pow(89) - 1), None),
             Primality::Probable(_)
         ));
+        assert!(matches!(
+            pb.is_prime(&Mint::from(2u128.pow(89) - 1), None),
+            Primality::Probable(_)
+        ));
 
         // test against small prime assertion
         for _ in 0..100 {
@@ -537,12 +576,27 @@ mod tests {
         }
 
         // test large numbers
+        const P: u128 = 18699199384836356663; // https://golang.org/issue/638
+        assert!(matches!(
+            pb.is_prime(&P, None),
+            Primality::Probable(_)
+        ));
+        assert!(matches!(
+            pb.is_prime(&P, Some(PrimalityTestConfig::bpsw())),
+            Primality::Probable(_)
+        ));
+        assert!(matches!(
+            pb.is_prime(&Mint::from(P), None),
+            Primality::Probable(_)
+        ));
+        assert!(matches!(
+            pb.is_prime(&Mint::from(P), Some(PrimalityTestConfig::bpsw())),
+            Primality::Probable(_)
+        ));
+
         #[cfg(feature = "num-bigint")]
         {
             let large_primes = [
-                // https://golang.org/issue/638
-                "18699199384836356663",
-
                 "98920366548084643601728869055592650835572950932266967461790948584315647051443",
                 "94560208308847015747498523884063394671606671904944666360068158221458669711639",
 
@@ -557,6 +611,9 @@ mod tests {
                 "9850501549098619803069760025035903451269934817616361666987073351061430442874302652853566563721228910201656997576599",                                           // E-382: 2^382-105
                 "42307582002575910332922579714097346549017899709713998034217522897561970639123926132812109468141778230245837569601494931472367",                                 // Curve41417: 2^414-17
                 "6864797660130609714981900799081393217269435300143305409394463459185543183397656052122559640661454554977296311391480858037121987999716643812574028291115057151", // E-521: 2^521-1
+
+                // https://github.com/AtropineTears/num-primes/issues/1#issuecomment-934629597
+                "169511182982703321453314585423962898651587669459838234386506572286328885534468792292646838949809616446341407457141008401355628947670484184607678853094537849610289912805960069455687743151708433319901176932959509872662610091644590437761688516626993416011399330087939042347256922771590903190536793274742859624657"
             ];
             for pstr in large_primes {
                 assert!(
