@@ -2,13 +2,17 @@
 
 use crate::traits::ExactRoots;
 use num_integer::{Integer, Roots};
-use num_modular::ModularOps;
+use num_modular::{ModularCoreOps, ModularUnaryOps};
 use num_traits::{FromPrimitive, NumRef, RefNum};
 use std::collections::BTreeMap;
 
-/// Find factors by trial division. The target is guaranteed fully factored
-/// only if bound() * bound() > target. The parameter limit sets the max prime to be tried aside from bound()
-/// Return the found factors and the residual. The residual will be Ok(1) or Ok(p) if fully factored.
+/// Find factors by trial division, returns a tuple of the found factors and the residual.
+///
+/// The target is guaranteed fully factored only if bound * bound > target, where bound = max(primes).
+/// The parameter limit additionally sets the maximum of primes to be tried.
+/// The residual will be Ok(1) or Ok(p) if fully factored.
+/// 
+/// TODO: implement fast check for small primes with BigInts in the precomputed table, and skip them in this function
 pub fn trial_division<
     I: Iterator<Item = u64>,
     T: Integer + Clone + Roots + NumRef + FromPrimitive,
@@ -55,26 +59,35 @@ where
     }
 }
 
-pub fn pollard_rho<T: Integer + FromPrimitive + NumRef + Clone>(
+/// Find factors using Pollard's rho algorithm with Brent's loop detection algorithm
+pub fn pollard_rho<
+    T: Integer
+        + FromPrimitive
+        + NumRef
+        + Clone
+        + for<'r> ModularCoreOps<&'r T, &'r T, Output = T>
+        + for<'r> ModularUnaryOps<&'r T, Output = T>,
+>(
     target: &T,
     start: T,
     offset: T,
 ) -> Option<T>
 where
-    for<'r> &'r T: RefNum<T> + ModularOps<&'r T, &'r T, Output = T>,
+    for<'r> &'r T: RefNum<T>,
 {
     let mut a = start.clone();
     let mut b = start;
-    // marker for loop detection, i = tortoise, j = hare
-    // see https://www.cnblogs.com/812-xiao-wen/p/10544546.html
+    // using Brent's loop detection, i = tortoise, j = hare
+    // XXX: further optimization see https://www.cnblogs.com/812-xiao-wen/p/10544546.html
     let (mut i, mut j) = (1usize, 2usize);
     while i > 0 {
         i += 1;
-        a = (&a).mulm(&a, &target).addm(&offset, &target);
+        a = a.sqm(&target).addm(&offset, &target);
         if a == b {
             return None;
         }
 
+        // FIXME: optimize abs_diff for montgomery form if we are going to use the abs_diff in the std lib
         let diff = if b > a { &b - &a } else { &a - &b }; // abs_diff
         let d = diff.gcd(target);
         if d > T::one() && &d < target {
@@ -92,10 +105,10 @@ where
 
 /// This function implements Shanks's square forms factorization (SQUFOF). It will assume that target
 /// is not a perfect square and the multiplier is square-free.
-pub fn squfof<T: Integer + NumRef + Clone + ExactRoots + std::fmt::Debug>(
-    target: &T,
-    multiplier: T,
-) -> Option<T>
+/// 
+/// Note that the multiplier is usually selected from the following sequence:
+/// `[1, 3, 5, 7, 11, 3*5, 3*7, 3*11, .., 3*5*7*11]`
+pub fn squfof<T: Integer + NumRef + Clone + ExactRoots>(target: &T, multiplier: T) -> Option<T>
 where
     for<'r> &'r T: RefNum<T>,
 {
@@ -168,22 +181,39 @@ where
     }
 }
 
-// TODO: ECM, Quadratic sieve / Prime field sieve, Fermat(https://en.wikipedia.org/wiki/Fermat%27s_factorization_method)
+// TODO: ECM, One line, Quadratic sieve / Prime field sieve, Fermat(https://en.wikipedia.org/wiki/Fermat%27s_factorization_method)
 // REF: https://pypi.org/project/primefac/
-fn pollard_brent() {}
+//      http://flintlib.org/doc/ulong_extras.html#factorisation
+//      https://github.com/zademn/facto-rs/
+//      https://github.com/elmomoilanen/prime-factorization
+//      https://cseweb.ucsd.edu/~ethome/teaching/2022-cse-291-14/
 fn pollard_pp1() {}
 fn williams_pp1() {}
 
 #[cfg(test)]
 mod tests {
     use super::*;
+    use crate::mint::Mint;
+    use num_modular::MontgomeryInt;
     use rand::random;
 
     #[test]
     fn pollard_rho_test() {
-        assert!(matches!(pollard_rho(&8051u16, 2, 1), Some(97)));
+        assert_eq!(pollard_rho(&8051u16, 2, 1), Some(97));
         assert!(matches!(pollard_rho(&8051u16, random(), 1), Some(i) if i == 97 || i == 83));
-        assert!(matches!(pollard_rho(&455459u32, 2, 1), Some(743)))
+        assert_eq!(pollard_rho(&455459u32, 2, 1), Some(743));
+
+        // Mint test
+        for _ in 0..10 {
+            let target = random::<u16>() | 1;
+            let start = random::<u16>() % target;
+            let offset = random::<u16>() % target;
+            assert_eq!(pollard_rho(&target, start, offset),
+                       pollard_rho(&Mint::from(target),
+                       MontgomeryInt::new(start, target).into(),
+                       MontgomeryInt::new(offset, target).into()
+                ).map(|v| v.value()));
+        }
     }
 
     #[test]
