@@ -12,21 +12,24 @@
 //!
 
 use crate::buffer::{NaiveBuffer, PrimeBufferExt};
+use crate::mint::Mint;
 use crate::factor::{pollard_rho, squfof};
 use crate::primality::{PrimalityBase, PrimalityRefBase};
-use crate::tables::{MOEBIUS_ODD, SMALL_PRIMES, WHEEL_NEXT, WHEEL_PREV, WHEEL_SIZE};
+use crate::tables::{MOEBIUS_ODD, SMALL_PRIMES, SMALL_PRIMES_NEXT, WHEEL_NEXT, WHEEL_PREV, WHEEL_SIZE};
 #[cfg(feature = "big-table")]
-use crate::tables::{SMALL_PRIMES_INV, SMALL_PRIMES_INVLIM, ZETA_LOG_TABLE};
+use crate::tables::{SMALL_PRIMES_INV, ZETA_LOG_TABLE};
 use crate::traits::{FactorizationConfig, Primality, PrimalityTestConfig, PrimalityUtils};
-use crate::RandPrime;
+use crate::{RandPrime, ExactRoots};
 #[cfg(feature = "num-bigint")]
 use num_bigint::{BigUint, RandBigInt};
 use num_integer::Roots;
-use num_modular::ModularOps;
+use num_modular::{ModularCoreOps, MontgomeryInt, ModularInteger};
+#[cfg(feature = "num-bigint")]
+use num_modular::DivExact;
 use num_traits::{CheckedAdd, FromPrimitive, Num, RefNum, ToPrimitive};
 use rand::{random, Rng};
 use std::collections::BTreeMap;
-use std::convert::TryFrom;
+use std::convert::{TryFrom, TryInto};
 
 #[cfg(feature = "big-table")]
 use crate::tables::{MILLER_RABIN_BASE32, MILLER_RABIN_BASE64};
@@ -37,66 +40,100 @@ use crate::tables::{MILLER_RABIN_BASE32, MILLER_RABIN_BASE64};
 #[cfg(not(feature = "big-table"))]
 pub fn is_prime64(target: u64) -> bool {
     // shortcuts
-    if target < 1 {
+    if target < 2 {
         return false;
     }
     if target & 1 == 0 {
         return target == 2;
-    }
-
-    // first find in the prime list
+    } 
     if let Ok(u) = u8::try_from(target) {
+        // find in the prime list if the target is small enough
         return SMALL_PRIMES.binary_search(&u).is_ok();
+    } else {
+        // check remainder against the wheel table
+        // this step eliminates any number that is not coprime to WHEEL_SIZE
+        let pos = (target % WHEEL_SIZE as u64) as usize;
+        if pos == 0 || WHEEL_NEXT[pos] < WHEEL_NEXT[pos-1] {
+            return false;
+        }
     }
 
     // Then do a deterministic Miller-rabin test
+    is_prime64_miller(target)
+}
+
+// Primality test for u64 with only miller-rabin tests, used during factorization.
+// It assumes the target is odd, not too small and cannot be divided small primes
+#[cfg(not(feature = "big-table"))]
+fn is_prime64_miller(target: u64) -> bool {
     // The collection of witnesses are from http://miller-rabin.appspot.com/
     if let Ok(u) = u16::try_from(target) {
         // 2, 3 for u16 range
-        return u.is_sprp(2) && u.is_sprp(3);
+        let u = Mint::from(u);
+        return u.is_sprp(Mint::from(2)) && u.is_sprp(Mint::from(3));
     }
     if let Ok(u) = u32::try_from(target) {
         // 2, 7, 61 for u32 range
-        return u.is_sprp(2) && u.is_sprp(7) && u.is_sprp(61);
+        let u = Mint::from(u);
+        return u.is_sprp(Mint::from(2)) && u.is_sprp(Mint::from(7)) && u.is_sprp(Mint::from(61));
     }
 
     // 2, 325, 9375, 28178, 450775, 9780504, 1795265022 for u64 range
     const WITNESS64: [u64; 7] = [2, 325, 9375, 28178, 450775, 9780504, 1795265022];
-    WITNESS64.iter().all(|&x| target.is_sprp(x))
+    let u = Mint::from(target);
+    WITNESS64.iter().all(|&x| u.is_sprp(Mint::from(x)))
 }
 
 /// Very fast primality test on a u64 integer is a prime number. It's based on
 /// deterministic Miller-rabin tests with hashing. if target is larger than 2^64 or more controlled
-/// primality tests are desired, please use is_prime() or PrimeBuffer::is_prime()
+/// primality tests are desired, please use [is_prime()]
 #[cfg(feature = "big-table")]
 pub fn is_prime64(target: u64) -> bool {
     // shortcuts
-    if target < 1 {
+    if target < 2 {
         return false;
     }
     if target & 1 == 0 {
         return target == 2;
     }
 
-    // first find in the prime list
-    if target < 8167 {
+    // trial division
+    if target < SMALL_PRIMES_NEXT {
+        // find in the prime list if the target is small enough
         return SMALL_PRIMES.binary_search(&(target as u16)).is_ok();
+    } else {
+        // check remainder against the wheel table
+        // this step eliminates any number that is not coprime to WHEEL_SIZE
+        let pos = (target % WHEEL_SIZE as u64) as usize;
+        if pos == 0 || WHEEL_NEXT[pos] < WHEEL_NEXT[pos-1] {
+            return false;
+        }
     }
 
+    is_prime64_miller(target)
+}
+
+// Primality test for u64 with only miller-rabin tests, used during factorization.
+// It assumes the target is odd, not too small and cannot be divided small primes
+#[cfg(feature = "big-table")]
+fn is_prime64_miller(target: u64) -> bool {
     // 32bit test
     const MAGIC: u32 = 0xAD625B89;
     if let Ok(u) = u32::try_from(target) {
         let base = u.wrapping_mul(MAGIC) >> 24;
-        return u.is_sprp(MILLER_RABIN_BASE32[base as usize]);
+        let u = Mint::from(u);
+        return u.is_sprp(Mint::from(MILLER_RABIN_BASE32[base as usize] as u32));
     }
 
     // 49bit test
-    if !target.is_sprp(2) {
+    let mt = Mint::from(target);
+    if !mt.is_sprp(2.into()) {
         return false;
     }
     let u = target as u32; // truncate
+
     let base = u.wrapping_mul(MAGIC) >> 18;
-    if !target.is_sprp(MILLER_RABIN_BASE64[base as usize] as u64) {
+    if !mt.is_sprp(Mint::from(MILLER_RABIN_BASE64[base as usize] as u64)) {
         return false;
     }
     if target < (1u64 << 49) {
@@ -106,15 +143,16 @@ pub fn is_prime64(target: u64) -> bool {
     // 64bit test
     const SECOND_BASES: [u64; 8] = [15, 135, 13, 60, 15, 117, 65, 29];
     let base = base >> 13;
-    target.is_sprp(SECOND_BASES[base as usize])
+    mt.is_sprp(Mint::from(SECOND_BASES[base as usize]))
 }
 
 /// Fast integer factorization on a u64 target. It's based on pollard's rho method and SQUFOF.
-/// if target is larger than 2^64 or more controlled primality tests are desired, please use [is_prime()].
+/// if target is larger than 2^64 or more controlled primality tests are desired, please use [factors()].
 ///
 /// The factorization can be quite faster under 2^64 because: 1) faster and deterministic primality check,
 /// 2) efficient montgomery multiplication implementation of u64
-pub fn factors64(target: u64) -> BTreeMap<u64, usize> {
+// TODO(v0.next): return BTreeMap<u64, u8> instead of BTreeMap<u64, usize>
+pub fn factorize64(target: u64) -> BTreeMap<u64, usize> {
     // TODO: improve factorization performance
     // REF: http://flintlib.org/doc/ulong_extras.html#factorisation
     //      https://mathoverflow.net/questions/114018/fastest-way-to-factor-integers-260
@@ -125,7 +163,9 @@ pub fn factors64(target: u64) -> BTreeMap<u64, usize> {
     //      https://github.com/radii/msieve
     //      Pari/GP: ifac_crack
     let mut result = BTreeMap::new();
-    let f2 = target.trailing_zeros(); // quick check on factors of 2
+
+    // quick check on factors of 2
+    let f2 = target.trailing_zeros();
     if f2 == 0 {
         if is_prime64(target) {
             result.insert(target, 1);
@@ -142,16 +182,15 @@ pub fn factors64(target: u64) -> BTreeMap<u64, usize> {
     let mut factored = false;
 
     #[cfg(not(feature = "big-table"))]
-    for &p in SMALL_PRIMES.iter().skip(1) {
-        let p64 = p as u64;
-        if p64 > tsqrt {
+    for p in SMALL_PRIMES.iter().skip(1).map(|&v| v as u64) {
+        if p > tsqrt {
             factored = true;
             break;
         }
 
-        while residual % p64 == 0 {
-            residual = residual / p64;
-            *result.entry(p64).or_insert(0) += 1;
+        while residual % p == 0 {
+            residual = residual / p;
+            *result.entry(p).or_insert(0) += 1;
         }
         if residual == 1 {
             factored = true;
@@ -160,34 +199,27 @@ pub fn factors64(target: u64) -> BTreeMap<u64, usize> {
     }
 
     #[cfg(feature = "big-table")]
-    // divisibility check with pre-computed tables
-    for (&p, (&pinv, plim)) in SMALL_PRIMES
-        .iter()
-        .zip(SMALL_PRIMES_INV.iter().zip(SMALL_PRIMES_INVLIM))
+    // divisibility check with pre-computed tables, see comments on SMALL_PRIMES_INV for reference 
+    for (p, &pinv) in SMALL_PRIMES
+        .iter().map(|&p| p as u64)
+        .zip(SMALL_PRIMES_INV.iter())
         .skip(1)
     {
-        let p64 = p as u64;
-        if p64 > tsqrt {
+        // only need to test primes up to sqrt(target)
+        if p > tsqrt {
             factored = true;
             break;
         }
 
-        let mut r = residual;
-        let mut k: u32 = 0;
-        loop {
-            let r2 = r.wrapping_mul(pinv);
+        let mut exp: usize = 0;
+        while let Some(q) = residual.div_exact(p, &pinv) {
+            exp += 1;
+            residual = q;
+        }
+        if exp > 0 {
+            result.insert(p, exp);
+        }
 
-            if r2 <= plim {
-                k += 1;
-                r = r2;
-            } else {
-                break;
-            }
-        }
-        if k > 0 {
-            residual = residual / p64.pow(k);
-            result.insert(p64, k as usize);
-        }
         if residual == 1 {
             factored = true;
             break;
@@ -201,8 +233,18 @@ pub fn factors64(target: u64) -> BTreeMap<u64, usize> {
         return result;
     }
 
-    // then try pollard's rho and SQUFOF methods util fully factored
-    let mut todo = vec![residual];
+    // then try advanced methods to find a divisor util fully factored
+    for (p, exp) in factorize64_advance(vec![(residual, 1usize)]).into_iter() {
+        *result.entry(p).or_insert(0) += exp;
+    }
+    result
+}
+
+// This function factorize all cofactors after some trivial division steps
+pub(crate) fn factorize64_advance(cofactors: Vec<(u64, usize)>) -> Vec<(u64, usize)> {
+    let mut todo = cofactors;
+    let mut factored: Vec<(u64, usize)> = Vec::new(); // prime factor, exponent
+
     const SQUFOF_MULTIPLIERS: [u16; 16] = [
         1,
         3,
@@ -221,35 +263,161 @@ pub fn factors64(target: u64) -> BTreeMap<u64, usize> {
         5 * 7 * 11,
         3 * 5 * 7 * 11,
     ];
-    while let Some(target) = todo.pop() {
-        if is_prime64(target) {
-            *result.entry(target).or_insert(0) += 1;
-        } else {
-            let mut i = 1usize;
-            let divisor = loop {
-                // try SQUFOF after 4 failed pollard rho trials
-                if i % 5 == 0 && (i / 5) < SQUFOF_MULTIPLIERS.len() {
-                    // TODO: check if the residual is a sqaure number before SQUFOF
-                    if let Some(p) = squfof(&target, SQUFOF_MULTIPLIERS[i / 5] as u64) {
-                        break p;
-                    }
-                } else {
-                    let start = random::<u64>() % target;
-                    let offset = random::<u64>() % target;
-                    if let Some(p) = pollard_rho(&target, start, offset) {
-                        break p;
-                    }
-                }
-                i += 1;
-            };
-            todo.push(divisor);
-            todo.push(target / divisor);
+    while let Some((target, exp)) = todo.pop() {
+        if is_prime64_miller(target) {
+            factored.push((target, exp));
+            continue;
         }
+
+        // check perfect powers before other methods
+        // it suffices to check square and cubic if big-table is enabled, since fifth power of
+        // the smallest prime that haven't been checked is 8167^5 > 2^64
+        if let Some(d) = target.sqrt_exact() {
+            todo.push((d, exp * 2));
+            continue;
+        }
+        if let Some(d) = target.cbrt_exact() {
+            todo.push((d, exp * 3));
+            continue;
+        }
+
+        // try to find a divisor
+        let mut i = 1usize;
+        let divisor = loop {
+            // try SQUFOF after 4 failed pollard rho trials
+            if i % 5 == 0 && (i / 5) < SQUFOF_MULTIPLIERS.len() {
+                if let Some(p) = squfof(&target, SQUFOF_MULTIPLIERS[i / 5] as u64) {
+                    break p;
+                }
+            } else {
+                let start = MontgomeryInt::new(random::<u64>(), target);
+                let offset = start.convert(random::<u64>());
+                if let Some(p) = pollard_rho(&Mint::from(target), start.into(), offset.into()) {
+                    break p.value();
+                }
+            }
+            i += 1;
+        };
+        todo.push((divisor, exp));
+        todo.push((target / divisor, exp));
+    }
+    factored
+}
+
+pub fn factorize128(target: u128) -> BTreeMap<u128, usize> {
+    // shortcut for u64
+    if target < (1u128 << 64) {
+        return factorize64(target as u64)
+            .into_iter()
+            .map(|(k, v)| (k as u128, v))
+            .collect();
+    }
+
+    let mut result = BTreeMap::new();
+
+    // quick check on factors of 2
+    let f2 = target.trailing_zeros();
+    if f2 != 0 {
+        result.insert(2, f2 as usize);
+    }
+    let mut residual = target >> f2;
+
+    // trial division using primes in the table
+    // note that p^2 is never larger than target, so we don't need to shortcut trial division
+    #[cfg(not(feature = "big-table"))]
+    for p in SMALL_PRIMES.iter().skip(1).map(|&v| v as u128) {
+        while residual % p == 0 {
+            residual = residual / p;
+            *result.entry(p).or_insert(0) += 1;
+        }
+        if residual == 1 {
+            return result;
+        }
+    }
+
+    #[cfg(feature = "big-table")]
+    // divisibility check with pre-computed tables, see comments on SMALL_PRIMES_INV for reference 
+    for (p, &pinv) in SMALL_PRIMES
+        .iter().map(|&p| p as u64)
+        .zip(SMALL_PRIMES_INV.iter())
+        .skip(1)
+    {
+        let mut exp: usize = 0;
+        while let Some(q) = residual.div_exact(p, &pinv) {
+            exp += 1;
+            residual = q;
+        }
+        if exp > 0 {
+            result.insert(p as u128, exp);
+        }
+
+        if residual == 1 {
+            return result;
+        }
+    }
+
+    // then try advanced methods to find a divisor util fully factored
+    let (mut todo128, mut todo64) = if let Ok(r64) = u64::try_from(residual) {
+        (Vec::new(), vec![(r64, 1usize)])
+    } else {
+        (vec![(residual, 1usize)], Vec::new())
+    }; // cofactors to be processed
+
+    while let Some((target, exp)) = todo128.pop() {
+        if is_prime(&Mint::from(target), None).probably() {
+            *result.entry(target).or_insert(0) += 1;
+            continue;
+        }
+
+        // check perfect powers before other methods
+        // it suffices to check 2, 3, 5, 7 power if big-table is enabled, since tenth power of
+        // the smallest prime that haven't been checked is 8167^10 > 2^128
+        if let Some(d) = target.sqrt_exact() {
+            if let Ok(d64) = u64::try_from(d) {
+                todo64.push((d64, exp * 2));
+            } else {
+                todo128.push((d, exp * 2));
+            }
+            continue;
+        }
+        if let Some(d) = target.cbrt_exact() {
+            if let Ok(d64) = u64::try_from(d) {
+                todo64.push((d64, exp * 3));
+            } else {
+                todo128.push((d, exp * 3));
+            }
+            continue;
+        }
+
+        let divisor = loop {
+            // TODO: only pollard rho is used by now, select better methods
+            let start = MontgomeryInt::new(random::<u128>(), target);
+            let offset = start.convert(random::<u128>());
+            if let Some(p) = pollard_rho(&Mint::from(target), start.into(), offset.into()) {
+                break p.value();
+            }
+        };
+
+        if let Ok(d64) = u64::try_from(divisor) {
+            todo64.push((d64, exp));
+        } else {
+            todo128.push((divisor, exp));
+        }
+        let co = target / divisor;
+        if let Ok(d64) = u64::try_from(co) {
+            todo64.push((d64, exp));
+        } else {
+            todo128.push((co, exp));
+        }
+    }
+
+    for (p, exp) in factorize64_advance(todo64).into_iter() {
+        *result.entry(p as u128).or_insert(0) += exp;
     }
     result
 }
 
-/// This function re-exports [crate::buffer::PrimeBufferExt::is_prime()] with a default buffer distance
+/// This function re-exports [PrimeBufferExt::is_prime()][crate::buffer::PrimeBufferExt::is_prime()] with a default buffer distance
 pub fn is_prime<T: PrimalityBase>(target: &T, config: Option<PrimalityTestConfig>) -> Primality
 where
     for<'r> &'r T: PrimalityRefBase<T>,
@@ -257,7 +425,7 @@ where
     NaiveBuffer::new().is_prime(target, config)
 }
 
-/// This function re-exports [crate::buffer::PrimeBufferExt::factors()] with a default buffer instance
+/// This function re-exports [PrimeBufferExt::factors()][crate::buffer::PrimeBufferExt::factors()] with a default buffer instance
 pub fn factors<T: PrimalityBase>(
     target: T,
     config: Option<FactorizationConfig>,
@@ -268,6 +436,14 @@ where
     NaiveBuffer::new().factors(target, config)
 }
 
+/// This function re-exports [PrimeBufferExt::factorize()][crate::buffer::PrimeBufferExt::factorize()] with a default buffer instance
+pub fn factorize<T: PrimalityBase>(target: T) -> BTreeMap<T, usize>
+where
+    for<'r> &'r T: PrimalityRefBase<T>,
+{
+    NaiveBuffer::new().factorize(target)
+}
+
 /// This function re-exports [NaiveBuffer::primes()] and collect result as a vector.
 pub fn primes(limit: u64) -> Vec<u64> {
     NaiveBuffer::new().into_primes(limit).collect()
@@ -296,14 +472,16 @@ pub fn primorial<T: PrimalityBase + std::iter::Product>(n: usize) -> T {
         .product()
 }
 
-/// This function calculate the Möbius μ(n) function of the input integer
+/// This function calculate the Möbius `μ(n)` function of the input integer `n`
 ///
+/// This function behaves like `moebius_factorized(factors(target, None).unwrap())`.
 /// If the input integer is very hard to factorize, it's better to use
-/// the [factors()] function to control how the factorization is done.
+/// the [factors()] function to control how the factorization is done, and then call
+/// [moebius_factorized()].
 ///
 /// # Panics
 /// if the factorization failed on target.
-pub fn moebius_mu<T: PrimalityBase>(target: &T) -> i8
+pub fn moebius<T: PrimalityBase>(target: &T) -> i8
 where
     for<'r> &'r T: PrimalityRefBase<T>,
 {
@@ -314,7 +492,7 @@ where
         if (target % four).is_zero() {
             return 0;
         } else {
-            return -moebius_mu(&(target / &two));
+            return -moebius(&(target / &two));
         }
     }
 
@@ -338,15 +516,9 @@ where
     }
 
     // then try complete factorization
-    let config = Some(FactorizationConfig::strict());
-    match factors(target.clone(), config) {
+    match factors(target.clone(), None) {
         Ok(result) => {
-            for exp in result.values() {
-                if exp > &1 {
-                    return 0;
-                }
-            }
-            return if result.len() % 2 == 0 { 1 } else { -1 };
+            return moebius_factorized(&result);
         }
         Err(_) => {
             panic!("Failed to factor the integer!");
@@ -354,6 +526,18 @@ where
     }
 }
 
+/// This function calculate the Möbius `μ(n)` function given the factorization
+/// result of `n`
+pub fn moebius_factorized<T>(factors: &BTreeMap<T, usize>) -> i8 {
+    if factors.values().any(|exp| exp > &1) {
+        0
+    } else if factors.len() % 2 == 0 {
+        1
+    } else {
+        -1
+    }
+}
+
 /// Tests if the integer doesn't have any square number factor.
 ///
 /// # Panics
@@ -362,7 +546,7 @@ pub fn is_square_free<T: PrimalityBase>(target: &T) -> bool
 where
     for<'r> &'r T: PrimalityRefBase<T>,
 {
-    moebius_mu(target) != 0
+    moebius(target) != 0
 }
 
 /// Returns the estimated bounds (low, high) of prime π function, such that
@@ -524,14 +708,16 @@ pub fn nth_prime_bounds<T: ToPrimitive + FromPrimitive>(target: &T) -> Option<(T
     }
 }
 
-/// Test if the target is a safe prime with Sophie German's definition. It will use the
-/// strict primality test configuration.
+/// Test if the target is a safe prime under [Sophie German's definition](https://en.wikipedia.org/wiki/Safe_and_Sophie_Germain_primes). It will use the
+/// [strict primality test configuration][FactorizationConfig::strict()].
 pub fn is_safe_prime<T: PrimalityBase>(target: &T) -> Primality
 where
     for<'r> &'r T: PrimalityRefBase<T>,
 {
     let buf = NaiveBuffer::new();
     let config = Some(PrimalityTestConfig::strict());
+    // XXX: use miller-rabin for large numbers (more than 256 bits?), as BPSW could be too slow (need check)
+    //      the NIST recommends 5 rounds for 512 and 1024 bits. For 1536 bits, the recommendation is 4 rounds.
 
     // test (n-1)/2 first since its smaller
     let sophie_p = buf.is_prime(&(target >> 1), config);
@@ -556,11 +742,14 @@ where
 {
     // first search in small primes
     if let Some(x) = target.to_u8() {
-        let next = match SMALL_PRIMES.binary_search(&target.to_u8().unwrap()) {
-            Ok(pos) => SMALL_PRIMES[pos + 1],
-            Err(pos) => SMALL_PRIMES[pos],
+        return match SMALL_PRIMES.binary_search(&x) {
+            Ok(pos) => if pos + 1 == SMALL_PRIMES.len() {
+                T::from_u64(SMALL_PRIMES_NEXT)
+            } else {
+                T::from_u8(SMALL_PRIMES[pos + 1])
+            },
+            Err(pos) => T::from_u8(SMALL_PRIMES[pos]),
         };
-        return T::from_u8(next);
     }
 
     // then moving along the wheel
@@ -587,7 +776,7 @@ where
     for<'r> &'r T: PrimalityRefBase<T>,
 {
     // first search in small primes
-    if target <= &T::from_u8(255).unwrap() // shortcut for u8
+    if target <= &T::from_u8(255).unwrap() // shortcut for T=u8
         || target < &T::from_u16(*SMALL_PRIMES.last().unwrap()).unwrap()
     {
         let next = match SMALL_PRIMES.binary_search(&target.to_u16().unwrap()) {
@@ -691,7 +880,7 @@ macro_rules! impl_randprime_prim {
                     panic!("The given bit size limit exceeded the capacity of the integer type!")
                 }
                 let t: $T = self.gen();
-                let t = t >> (<$T>::BITS - bit_size as u32);
+                let t = (t >> (<$T>::BITS - bit_size as u32)) | 1; // filter even numbers
                 if is_prime64(t as u64) {
                     t
                 } else {
@@ -734,7 +923,7 @@ impl<R: Rng> RandPrime<u128> for R {
             panic!("The given bit size limit exceeded the capacity of the integer type!")
         }
         let t: u128 = self.gen();
-        let t = t >> (u128::BITS - bit_size as u32);
+        let t = (t >> (u128::BITS - bit_size as u32)) | 1; // filter even numbers
         if is_prime(&t, config).probably() {
             t
         } else {
@@ -762,11 +951,11 @@ impl<R: Rng> RandPrime<u128> for R {
     }
 }
 
-#[cfg(feature = "big-int")]
+#[cfg(feature = "num-bigint")]
 impl<R: Rng> RandPrime<BigUint> for R {
     #[inline]
     fn gen_prime(&mut self, bit_size: usize, config: Option<PrimalityTestConfig>) -> BigUint {
-        let t = self.gen_biguint(bit_size as u64);
+        let t = self.gen_biguint(bit_size as u64) | BigUint::from(1u8); // filter even numbers
         if is_prime(&t, config).probably() {
             t
         } else {
@@ -915,20 +1104,29 @@ mod tests {
     }
 
     #[test]
-    fn factors64_test() {
-        // some known cases
+    fn factorize64_test() {
+        // some simple cases
+        let fac4095 = BTreeMap::from_iter([(3, 2), (5, 1), (7, 1), (13, 1)]);
+        let fac = factorize64(4095);
+        assert_eq!(fac, fac4095);
+
         let fac123456789 = BTreeMap::from_iter([(3, 2), (3803, 1), (3607, 1)]);
-        let fac = factors64(123456789);
+        let fac = factorize64(123456789);
         assert_eq!(fac, fac123456789);
 
         let fac1_17 = BTreeMap::from_iter([(2071723, 1), (5363222357, 1)]);
-        let fac = factors64(11111111111111111);
+        let fac = factorize64(11111111111111111);
         assert_eq!(fac, fac1_17);
 
+        // perfect powers
+        for exp in 2u32..5 {
+            assert_eq!(factorize128(8167u128.pow(exp)), BTreeMap::from_iter([(8167, exp as usize)]));
+        }
+
         // 100 random factorization tests
         for _ in 0..100 {
             let x = random();
-            let fac = factors64(x);
+            let fac = factorize64(x);
             let mut prod = 1;
             for (p, exp) in fac {
                 assert!(
@@ -942,6 +1140,40 @@ mod tests {
         }
     }
 
+    #[test]
+    fn factorize128_test() {
+        // some simple cases
+        let fac_primorial19 = BTreeMap::from_iter(SMALL_PRIMES.iter().take(19).map(|&p| (p as u128, 1)));
+        let fac = factorize128(7858321551080267055879090);
+        assert_eq!(fac, fac_primorial19);
+        
+        let fac_smallbig = BTreeMap::from_iter([(167, 1), (2417851639229258349412369, 1)]);
+        let fac = factorize128(403781223751286144351865623);
+        assert_eq!(fac, fac_smallbig);
+
+        // perfect powers
+        for exp in 5u32..10 {
+            // 2^64 < 8167^5 < 8167^9 < 2^128
+            assert_eq!(factorize128(8167u128.pow(exp)), BTreeMap::from_iter([(8167, exp as usize)]));
+        }
+
+        // random factorization tests
+        for _ in 0..1 { // TODO: run more tests when other factorization methods are implemented
+            let x = random();
+            let fac = factorize128(x);
+            let mut prod = 1;
+            for (p, exp) in fac {
+                assert!(
+                    is_prime(&p, None).probably(),
+                    "factorization result should have prime factors! (get {})",
+                    p
+                );
+                prod *= p.pow(exp as u32);
+            }
+            assert_eq!(x, prod, "factorization check failed! ({} != {})", x, prod);
+        }
+    }
+
     #[test]
     fn is_safe_prime_test() {
         // OEIS:A005385
@@ -963,18 +1195,18 @@ mod tests {
     }
 
     #[test]
-    fn moebius_mu_test() {
+    fn moebius_test() {
         // test small examples
         let mu20: [i8; 20] = [
             1, -1, -1, 0, -1, 1, -1, 0, 0, 1, -1, 0, -1, 1, 1, 0, -1, 0, -1, 0,
         ];
         for i in 0..20 {
-            assert_eq!(moebius_mu(&(i + 1)), mu20[i], "moebius on {}", i);
+            assert_eq!(moebius(&(i + 1)), mu20[i], "moebius on {}", i);
         }
 
         // some square numbers
-        assert_eq!(moebius_mu(&1024u32), 0);
-        assert_eq!(moebius_mu(&(8081u32 * 8081)), 0);
+        assert_eq!(moebius(&1024u32), 0);
+        assert_eq!(moebius(&(8081u32 * 8081)), 0);
 
         // sphenic numbers
         let sphenic3: [u8; 20] = [
@@ -982,14 +1214,14 @@ mod tests {
             195, 222,
         ]; // OEIS:A007304
         for i in 0..20 {
-            assert_eq!(moebius_mu(&sphenic3[i]), -1i8, "moebius on {}", sphenic3[i]);
+            assert_eq!(moebius(&sphenic3[i]), -1i8, "moebius on {}", sphenic3[i]);
         }
         let sphenic5: [u16; 23] = [
             2310, 2730, 3570, 3990, 4290, 4830, 5610, 6006, 6090, 6270, 6510, 6630, 7410, 7590,
             7770, 7854, 8610, 8778, 8970, 9030, 9282, 9570, 9690,
         ]; // OEIS:A046387
         for i in 0..20 {
-            assert_eq!(moebius_mu(&sphenic5[i]), -1i8, "moebius on {}", sphenic5[i]);
+            assert_eq!(moebius(&sphenic5[i]), -1i8, "moebius on {}", sphenic5[i]);
         }
     }
 
@@ -1097,6 +1329,13 @@ mod tests {
     fn prev_next_test() {
         assert_eq!(prev_prime(&2u32, None), None);
 
+        // prime table boundary test
+        assert_eq!(prev_prime(&257u16, None), Some(251));
+        assert_eq!(next_prime(&251u16, None), Some(257));
+        assert_eq!(next_prime(&251u8, None), None);
+        assert_eq!(prev_prime(&8167u16, None), Some(8161));
+        assert_eq!(next_prime(&8161u16, None), Some(8167));
+
         // OEIS:A077800
         let twine_primes: [(u32, u32); 8] = [
             (2, 3), // not exactly twine
@@ -1156,12 +1395,12 @@ mod tests {
         let p: u128 = rng.gen_safe_prime(128);
         assert!(is_safe_prime(&p).probably());
 
-        #[cfg(feature = "big-int")]
+        #[cfg(feature = "num-bigint")]
         {
             let p: BigUint = rng.gen_prime(512, None);
             assert!(is_prime(&p, None).probably());
             let p: BigUint = rng.gen_safe_prime(192);
-            assert!(is_prime(&p, None).probably());
+            assert!(is_safe_prime(&p).probably());
         }
 
         // test bit size limit
