@@ -1,7 +1,7 @@
 use crate::traits::{BitTest, ExactRoots, PrimalityUtils};
 use either::Either;
 use num_integer::{Integer, Roots};
-use num_modular::ModularOps;
+use num_modular::{ModularCoreOps, ModularRefOps, ModularUnaryOps};
 use num_traits::{FromPrimitive, NumRef, RefNum, ToPrimitive};
 
 /// Utilities for the Lucas pseudoprime test
@@ -20,10 +20,18 @@ pub trait LucasUtils {
     fn p_bruteforce(n: &Self) -> usize;
 }
 
-impl<T: Integer + FromPrimitive + ToPrimitive + NumRef + BitTest + ExactRoots + Clone> LucasUtils
-    for T
+impl<T> LucasUtils for T
 where
-    for<'r> &'r T: RefNum<T> + ModularOps<&'r T, &'r T, Output = T>,
+    T: Integer
+        + FromPrimitive
+        + ToPrimitive
+        + NumRef
+        + BitTest
+        + ExactRoots
+        + Clone
+        + ModularRefOps,
+    for<'r> &'r T:
+        RefNum<T> + ModularUnaryOps<&'r T, Output = T> + ModularCoreOps<&'r T, &'r T, Output = T>,
 {
     fn lucasm(p: usize, q: isize, m: Self, n: Self) -> (Self, Self) {
         // Reference: <https://en.wikipedia.org/wiki/Lucas_sequence>
@@ -33,42 +41,44 @@ where
         let q = if q >= 0 {
             T::from_isize(q).unwrap() % &m
         } else {
-            (&T::from_isize(-q).unwrap()).negm(&m)
+            T::from_isize(-q).unwrap().negm(&m)
         };
 
-        let mut uk = T::zero(); // U(k)
-        let mut uk1 = T::one(); // U(k+1)
-        let two = T::one() + T::one();
+        let mut uk = T::zero() % &m; // U(k), mod m for montgomery form
+        let mut uk1 = T::one() % &m; // U(k+1)
 
         for i in (0..n.bits()).rev() {
             if n.bit(i) {
                 // k' = 2k+1
                 // U(k'+1) = U(2k+2) = PU(k+1)² - 2*QU(k+1)U(k)
-                let t1 = p.mulm(&uk1, &m).mulm(&uk1, &m);
-                let t2 = two.mulm(&q, &m).mulm(&uk1, &m).mulm(&uk, &m);
+                let uk1sq = (&uk1).sqm(&m);
+                let t1 = (&p).mulm(&uk1sq, &m);
+                let t2 = (&q).mulm(&uk1, &m).mulm(&uk, &m).dblm(&m);
                 let new_uk1 = t1.subm(&t2, &m);
                 // U(k') = U(2k+1) = U(k+1)² - QU(k)²
-                let t1 = uk1.mulm(&uk1, &m);
-                let t2 = q.mulm(&uk, &m).mulm(&uk, &m);
+                let t1 = uk1sq;
+                let t2 = (&q).mulm(&uk.sqm(&m), &m);
                 let new_uk = t1.subm(&t2, &m);
                 uk1 = new_uk1;
                 uk = new_uk;
             } else {
                 // k' = 2k
                 // U(k'+1) = U(2k+1) = U(k+1)² - QU(k)²
-                let t1 = uk1.mulm(&uk1, &m);
-                let t2 = q.mulm(&uk, &m).mulm(&uk, &m);
+                let uksq = (&uk).sqm(&m);
+                let t1 = (&uk1).sqm(&m);
+                let t2 = (&uksq).mulm(&q, &m);
                 let new_uk1 = t1.subm(&t2, &m);
                 // U(k') = U(2k) = 2U(k+1)U(k) - PU(k)²
-                let t1 = two.mulm(&uk1, &m).mulm(&uk, &m);
-                let t2 = p.mulm(&uk, &m).mulm(&uk, &m);
+                let t1 = uk1.mulm(&uk, &m).dblm(&m);
+                let t2 = uksq.mulm(&p, &m);
                 let new_uk = t1.subm(&t2, &m);
                 uk1 = new_uk1;
                 uk = new_uk;
             }
         }
 
-        let vk = two.mulm(&uk1, &m).subm(&p.mulm(&uk, &m), &m);
+        // V(k) = 2U(k+1) - PU(k)
+        let vk = (&uk1).dblm(&m).subm(&p.mulm(&uk, &m), &m);
         (uk, vk)
     }
 
@@ -82,7 +92,7 @@ where
             }
 
             let sd = if neg { (&d).negm(n) } else { d.clone() };
-            let j = ModularOps::<&T, &T>::jacobi(&sd, n);
+            let j = sd.jacobi(n);
             if j == 0 && &d != n {
                 break (0, 0);
             } // modification from Baillie, see https://oeis.org/A217120/a217120_1.txt
@@ -109,7 +119,7 @@ where
             }
 
             let d = T::from_usize(p * p - 4).unwrap();
-            let j = ModularOps::<&T, &T>::jacobi(&d, n);
+            let j = d.jacobi(n);
             if j == 0 && &d != n {
                 break 0;
             }
@@ -121,11 +131,13 @@ where
     }
 }
 
-impl<T: Integer + NumRef + Clone + FromPrimitive + LucasUtils + BitTest> PrimalityUtils for T
+impl<T: Integer + NumRef + Clone + FromPrimitive + LucasUtils + BitTest + ModularRefOps>
+    PrimalityUtils for T
 where
     for<'r> &'r T:
-        RefNum<T> + std::ops::Shr<usize, Output = T> + ModularOps<&'r T, &'r T, Output = T>,
+        RefNum<T> + std::ops::Shr<usize, Output = T> + ModularUnaryOps<&'r T, Output = T>,
 {
+    #[inline]
     fn is_prp(&self, base: Self) -> bool {
         if self < &Self::one() {
             return false;
@@ -134,6 +146,7 @@ where
         base.powm(&tm1, self).is_one()
     }
 
+    #[inline]
     fn is_sprp(&self, base: Self) -> bool {
         self.test_sprp(base).either(|v| v, |_| false)
     }
@@ -148,23 +161,27 @@ where
         let shift = tm1.trailing_zeros();
         let u = &tm1 >> shift;
 
+        // prevent reduction if the input is in montgomery form
+        let m1 = T::one() % self;
+        let mm1 = (&m1).negm(self);
+
         let mut x = base.powm(&u, self);
-        if x == T::one() || x == tm1 {
+        if x == m1 || x == mm1 {
             return Either::Left(true);
         }
 
         for _ in 0..shift {
-            let y = (&x).mulm(&x, self);
-            if y.is_one() {
+            let y = (&x).sqm(self);
+            if y == m1 {
                 return Either::Right(self.gcd(&(x - T::one())));
             }
-            if y == tm1 {
+            if y == mm1 {
                 return Either::Left(true);
             }
             x = y;
         }
 
-        Either::Left(x == T::one())
+        Either::Left(x == m1)
     }
 
     fn is_lprp(&self, p: Option<usize>, q: Option<isize>) -> bool {
@@ -190,13 +207,13 @@ where
         let d = if d > 0 {
             Self::from_isize(d).unwrap()
         } else {
-            (&Self::from_isize(-d).unwrap()).negm(self)
+            Self::from_isize(-d).unwrap().negm(self)
         };
         let delta = match d.jacobi(self) {
             0 => self.clone(),
             -1 => self + Self::one(),
             1 => self - Self::one(),
-            _ => panic!(""),
+            _ => unreachable!(),
         };
         let (u, _) = LucasUtils::lucasm(p, q, self.clone(), delta);
         u.is_zero()
@@ -225,13 +242,13 @@ where
         let d = if d > 0 {
             Self::from_isize(d).unwrap()
         } else {
-            (&Self::from_isize(-d).unwrap()).negm(self)
+            Self::from_isize(-d).unwrap().negm(self)
         };
         let delta = match d.jacobi(self) {
             0 => self.clone(),
             -1 => self + Self::one(),
             1 => self - Self::one(),
-            _ => panic!(""),
+            _ => unreachable!(),
         };
 
         let shift = delta.trailing_zeros();
@@ -247,25 +264,24 @@ where
             return false;
         }
         let qk = Self::from_isize(q.abs()).unwrap().powm(&base, self);
-        let two = Self::from_u8(2).unwrap();
         // V(2k) = V(k)² - 2Q^k
         let mut vd = if q >= 0 {
-            vd.mulm(&vd, self).subm(&two.mulm(&qk, self), self)
+            vd.sqm(self).subm(&(&qk).dblm(self), self)
         } else {
-            vd.mulm(&vd, self).addm(&two.mulm(&qk, self), self)
+            vd.sqm(self).addm(&(&qk).dblm(self), self)
         };
         if vd.is_zero() {
             return true;
         }
-        let mut qk = qk.mulm(&qk, self);
+        let mut qk = qk.sqm(self);
 
         for _ in 1..shift {
             // V(2k) = V(k)² - 2Q^k
-            vd = vd.mulm(&vd, self).subm(&two.mulm(&qk, self), self);
+            vd = vd.sqm(self).subm(&(&qk).dblm(self), self);
             if vd.is_zero() {
                 return true;
             }
-            qk = qk.mulm(&qk, self);
+            qk = qk.sqm(self);
         }
         false
     }
@@ -293,13 +309,13 @@ where
         let d = if d > 0 {
             Self::from_isize(d).unwrap()
         } else {
-            (&Self::from_isize(-d).unwrap()).negm(self)
+            Self::from_isize(-d).unwrap().negm(self)
         };
         let delta = match d.jacobi(self) {
             0 => self.clone(),
             -1 => self + Self::one(),
             1 => self - Self::one(),
-            _ => panic!(""),
+            _ => unreachable!(),
         };
 
         let shift = delta.trailing_zeros();
@@ -317,7 +333,7 @@ where
 
         for _ in 1..(shift - 1) {
             // V(2k) = V(k)² - 2
-            vd = vd.mulm(&vd, self).subm(&two, self);
+            vd = vd.sqm(self).subm(&two, self);
             if vd.is_zero() {
                 return true;
             }
@@ -329,11 +345,28 @@ where
 /// A dummy trait for integer type. All types that implements this and [PrimalityRefBase]
 /// will be supported by most functions in `num-primes`
 pub trait PrimalityBase:
-    Integer + Roots + NumRef + Clone + FromPrimitive + ToPrimitive + ExactRoots + BitTest
+    Integer
+    + Roots
+    + NumRef
+    + Clone
+    + FromPrimitive
+    + ToPrimitive
+    + ExactRoots
+    + BitTest
+    + ModularRefOps
 {
 }
-impl<T: Integer + Roots + NumRef + Clone + FromPrimitive + ToPrimitive + ExactRoots + BitTest>
-    PrimalityBase for T
+impl<
+        T: Integer
+            + Roots
+            + NumRef
+            + Clone
+            + FromPrimitive
+            + ToPrimitive
+            + ExactRoots
+            + BitTest
+            + ModularRefOps,
+    > PrimalityBase for T
 {
 }
 
@@ -342,19 +375,23 @@ impl<T: Integer + Roots + NumRef + Clone + FromPrimitive + ToPrimitive + ExactRo
 pub trait PrimalityRefBase<Base>:
     RefNum<Base>
     + std::ops::Shr<usize, Output = Base>
-    + for<'r> ModularOps<&'r Base, &'r Base, Output = Base>
+    + for<'r> ModularUnaryOps<&'r Base, Output = Base>
+    + for<'r> ModularCoreOps<&'r Base, &'r Base, Output = Base>
 {
 }
 impl<T, Base> PrimalityRefBase<Base> for T where
     T: RefNum<Base>
         + std::ops::Shr<usize, Output = Base>
-        + for<'r> ModularOps<&'r Base, &'r Base, Output = Base>
+        + for<'r> ModularUnaryOps<&'r Base, Output = Base>
+        + for<'r> ModularCoreOps<&'r Base, &'r Base, Output = Base>
 {
 }
 
 #[cfg(test)]
 mod tests {
     use super::*;
+    use crate::mint::Mint;
+    use num_modular::{ModularAbs, ModularSymbols};
     use rand::random;
 
     #[cfg(feature = "num-bigint")]
@@ -379,10 +416,12 @@ mod tests {
         let spsp: [u16; 5] = [2047, 3277, 4033, 4681, 8321];
         for psp in spsp {
             assert!(psp.is_sprp(2));
+            assert!(Mint::from(psp).is_sprp(2.into())); // test Mint execution
         }
 
         // test cofactor return
-        assert!(matches!(341u16.test_sprp(2), Either::Right(31)));
+        assert_eq!(341u16.test_sprp(2), Either::Right(31));
+        assert_eq!(Mint::from(341u16).test_sprp(2.into()), Either::Right(31.into()));
     }
 
     #[test]
@@ -427,6 +466,43 @@ mod tests {
                 assert_eq!(uk, BigUint::from(p3qm1[n] % (m as u64)));
             }
         }
+
+        fn lucasm_naive(p: usize, q: isize, m: u16, n: u16) -> (u16, u16) {
+            if n == 0 {
+                return (0, 2);
+            }
+
+            let m = m as usize;
+            let q = q.absm(&m);
+            let (mut um1, mut u) = (0, 1); // U_{n-1}, U_{n}
+            let (mut vm1, mut v) = (2, p % m); // V_{n-1}, V_{n}
+
+            for _ in 1..n {
+                let new_u = p.mulm(u, &m).subm(q.mulm(um1, &m), &m);
+                um1 = u;
+                u = new_u;
+
+                let new_v = p.mulm(v, &m).subm(q.mulm(vm1, &m), &m);
+                vm1 = v;
+                v = new_v;
+            }
+            (u as u16, v as u16)
+        }
+        for _ in 0..10 {
+            let n = random::<u8>() as u16;
+            let m = random::<u16>();
+            let p = random::<u16>() as usize;
+            let q = random::<i16>() as isize;
+            assert_eq!(
+                LucasUtils::lucasm(p, q, m, n),
+                lucasm_naive(p, q, m, n),
+                "failed with Lucas settings: p={}, q={}, m={}, n={}",
+                p,
+                q,
+                m,
+                n
+            );
+        }
     }
 
     #[test]
@@ -452,7 +528,7 @@ mod tests {
                     continue;
                 } // skip real primes
                 let d = (p * p + 4) as u16;
-                if n.is_odd() && ModularOps::<&u16, &u16>::jacobi(&d, &n) != -1 {
+                if n.is_odd() && d.jacobi(&n) != -1 {
                     continue;
                 }
                 assert!(
@@ -477,7 +553,7 @@ mod tests {
                     continue;
                 } // skip real primes
                 let d = (p * p + 4) as u16;
-                if n.is_odd() && ModularOps::<&u16, &u16>::jacobi(&d, &n) != -1 {
+                if n.is_odd() && d.jacobi(&n) != -1 {
                     continue;
                 }
                 assert!(
