@@ -1,3 +1,4 @@
+use core::default::Default;
 use std::ops::{BitAnd, BitOr};
 
 use either::Either;
@@ -17,7 +18,7 @@ pub trait BitTest {
 }
 
 /// This enum describes the result of primality checks
-#[derive(Debug, Clone, Copy)]
+#[derive(Debug, Clone, Copy, PartialEq)]
 pub enum Primality {
     /// The number passes deterministic primality check.
     Yes,
@@ -31,7 +32,8 @@ pub enum Primality {
 
 impl Primality {
     /// Check whether the resule indicates that the number is
-    /// (very) probably a prime. Return false only on Primaliy::No
+    /// (very) probably a prime. Return false only on [Primality::No]
+    #[inline(always)]
     pub fn probably(self) -> bool {
         match self {
             Primality::No => false,
@@ -43,7 +45,7 @@ impl Primality {
 impl BitAnd<Primality> for Primality {
     type Output = Primality;
 
-    /// Combine two primality result by ensuring both numbers are prime
+    /// Combine two primality results by ensuring both numbers are prime
     fn bitand(self, rhs: Primality) -> Self::Output {
         match self {
             Primality::No => Primality::No,
@@ -60,7 +62,7 @@ impl BitAnd<Primality> for Primality {
 impl BitOr<Primality> for Primality {
     type Output = Primality;
 
-    /// Combine two primality result by ensuring either numbers is prime
+    /// Combine two primality results by ensuring either numbers is prime
     fn bitor(self, rhs: Primality) -> Self::Output {
         match self {
             Primality::No => rhs,
@@ -76,7 +78,11 @@ impl BitOr<Primality> for Primality {
 
 /// Represents a configuration for a primality test
 #[derive(Debug, Clone, Copy)]
+#[non_exhaustive]
 pub struct PrimalityTestConfig {
+    // TODO: add option to divides small primes in the table
+    //       and this option should be enabled if the probabilistic test is used for strict config
+
     /// Number of strong probable prime test, starting from base 2
     pub sprp_trials: usize,
 
@@ -90,19 +96,21 @@ pub struct PrimalityTestConfig {
     pub eslprp_test: bool,
 }
 
-impl PrimalityTestConfig {
+impl Default for PrimalityTestConfig {
     /// Create a defalt primality testing configuration. This config will eliminate most
     /// composites with little computation
-    pub fn default() -> Self {
+    fn default() -> Self {
         Self {
-            sprp_trials: 2,
-            sprp_random_trials: 2,
+            sprp_trials: 2,        // test base 2 and 3
+            sprp_random_trials: 3, // choose other 3 random bases
             slprp_test: false,
             eslprp_test: false,
         }
     }
+}
 
-    /// Create a configuration with the known stongest deterministic primality test
+impl PrimalityTestConfig {
+    /// Create a configuration with the **stongest deterministic** primality test available
     pub fn strict() -> Self {
         Self::bpsw() // TODO: change to 2-base SPRP + VPRP
     }
@@ -125,9 +133,10 @@ impl PrimalityTestConfig {
 
 /// Represents a configuration for integer factorization
 #[derive(Debug, Clone, Copy)]
+#[non_exhaustive]
 pub struct FactorizationConfig {
     /// Config for testing if a factor is prime
-    pub prime_test_config: PrimalityTestConfig,
+    pub primality_config: PrimalityTestConfig,
 
     /// Prime limit of trial division, you also need to reserve the primes in the buffer
     /// if all primes under the limit are to be tested. `None` means using all available primes.
@@ -136,9 +145,6 @@ pub struct FactorizationConfig {
     /// Number of trials with Pollard's rho method
     pub rho_trials: usize,
 
-    /// Number of trials with Pollard's rho method (Brent variant)
-    brent_trials: usize,
-
     /// Number of trials with Pollard's p-1 method
     pm1_trials: usize,
 
@@ -146,25 +152,26 @@ pub struct FactorizationConfig {
     pp1_trials: usize,
 }
 
-impl FactorizationConfig {
+impl Default for FactorizationConfig {
     /// Create a defalt primality testing configuration. This config will factorize
     /// most integers within decent time
-    pub fn default() -> Self {
+    fn default() -> Self {
         const THRESHOLD_DEFAULT_TD: u64 = 1 << 14;
         Self {
-            prime_test_config: PrimalityTestConfig::default(),
+            primality_config: PrimalityTestConfig::default(),
             td_limit: Some(THRESHOLD_DEFAULT_TD),
             rho_trials: 4,
-            brent_trials: 0,
             pm1_trials: 0,
             pp1_trials: 0,
         }
     }
+}
 
+impl FactorizationConfig {
     /// Same as the default configuration but with strict primality check
     pub fn strict() -> Self {
         let mut config = Self::default();
-        config.prime_test_config = PrimalityTestConfig::strict();
+        config.primality_config = PrimalityTestConfig::strict();
         config
     }
 }
@@ -197,12 +204,8 @@ pub trait ExactRoots: Roots + Pow<u32, Output = Self> + Clone {
     }
 }
 
-// TODO: implement quick div_exact (which might be useful in various functions)
-// REF: GMP `mpz_divexact`
-//      FLINT `fmpz_divexact`
-//      factor.c `divexact_21`
-
-// TODO: implement quick is_x_power (specifically is_235_power)
+// TODO: implement is_perfect_power, this could be used during factorization
+//       to filter out perfect powers when factorize large integers
 // REF: PARI/GP `Z_ispowerall`, `is_357_power`
 //      FLINT `n_is_perfect_power235`, `fmpz_is_perfect_power`
 //      GMP `mpz_perfect_power_p`
@@ -273,4 +276,6 @@ pub trait RandPrime<T> {
     /// Generate a random (Sophie German) safe prime within the given bit size limit. The generated prime
     /// is guaranteed to pass the [is_safe_prime][crate::nt_funcs::is_safe_prime] test
     fn gen_safe_prime(&mut self, bit_size: usize) -> T;
+
+    // TODO: add gen_prime_exact and gen_safe_prime_exact for generating primes with exact this bit size
 }
